
// #include <cutil.h>



#define DEBUG  if(0)

#ifdef  __DEVICE_EMULATION__
#define USE_PRINTF
#endif

#ifdef USE_PRINTF
#define PRINTF(...)  printf(__VA_ARGS__)
#else
#define PRINTF(...) 
#endif


#define NOSERIALIZE(...)  __VA_ARGS__	


#ifdef  __DEVICE_EMULATION__

#define SERIALIZE(...)    __VA_ARGS__	

#else

#define SERIALIZE(...)			      \
  {					      \
    int i ;				      \
    for (i=0; i<blockDim.x; i++)	      \
      {					      \
	__syncthreads();		      \
	if (i == threadIdx.x)		      \
	  {  __VA_ARGS__  }		      \
	    }				      \
  }

#endif


#define CUDA_SAFE_CALL(call) {						\
    cudaError err = call;						\
    if(cudaSuccess != err) {						\
      fprintf(stderr, "Cuda error in file '%s' in line %i : %s.\n",	\
	      __FILE__, __LINE__, cudaGetErrorString( err) );		\
      exit(EXIT_FAILURE);						\
    } 


#  define CUT_CHECK_ERROR(errorMessage) do {                                 \
    cudaError_t err = cudaGetLastError();                                    \
    if( cudaSuccess != err) {                                                \
      fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
	      errorMessage, __FILE__, __LINE__, cudaGetErrorString( err) );\
      exit(EXIT_FAILURE);                                                  \
    }                                                                        \
    err = cudaThreadSynchronize();                                           \
    if( cudaSuccess != err) {                                                \
      fprintf(stderr, "Cuda error: %s in file '%s' in line %i : %s.\n",    \
	      errorMessage, __FILE__, __LINE__, cudaGetErrorString( err) );\
      exit(EXIT_FAILURE);                                                  \
    } } while (0)


