

void init_cuda()
{
  int deviceCount, dev;
  
  cudaGetDeviceCount(&deviceCount);
  
  if (deviceCount == 0)
    {
      perror("No CUDA device");
      exit(9);
    }
	
  printf("\n");

  for (dev = 0; dev < deviceCount; ++dev)
    {
      cudaDeviceProp deviceProp;
      cudaGetDeviceProperties(&deviceProp, dev);
      
      printf("==[%d]== %s\n", dev, deviceProp.name );

      if (conf_verbose)
      {
      printf("        cc %d.%d\n", deviceProp.major, deviceProp.minor);
      printf("%d kHz\n", deviceProp.clockRate);
      printf("Memory   global: %d bytes (%d MB)\n", deviceProp.totalGlobalMem, (long) deviceProp.totalGlobalMem / (1 << 30)); 
      printf("         const: %d bytes\n", deviceProp.totalConstMem); 
      printf("         shared : %d bytes / block\n", deviceProp.sharedMemPerBlock ); 
      printf("         regs: %d / block\n", deviceProp.regsPerBlock ); 
      printf("warpSize %d\n", deviceProp.warpSize );
      printf("Limits   threads: <%d, %d, %d>\n", deviceProp.maxThreadsDim[0], deviceProp.maxThreadsDim[1], deviceProp.maxThreadsDim[2]); 
      printf("         threads per block: %d\n", deviceProp.maxThreadsPerBlock); 
      printf("         grid: <%d, %d, %d>\n", deviceProp.maxGridSize[0], deviceProp.maxGridSize[1], deviceProp.maxGridSize[2]); 
      printf("         memPitch: %d\n", deviceProp.memPitch );
      }
      printf("\n");      
    }
  
  if (cudaSetDevice(0))
    {
      perror("CUDA Device init failed.");
      exit(4);
    }

#ifdef  __DEVICE_EMULATION__
  printf("==[x]== Using Device Emulation (CPU)\n\n");
#endif

}

