/*
 * CUDA PWM - Mathieu Giraud & Jean-Stéphane Varré, 2008
 */



#define PLOT_WIDTH  100
#define PLOT_HEIGTH 10 



hcount datap[PLOT_WIDTH] ;
int datan[PLOT_WIDTH] ;

void sprintf_largeint(char *buf, hcount x, int l)
{
  if (x < pow(10.0, (double) l))
    sprintf(buf, "%*d", l, (int) x);
        else
	        sprintf(buf, "%*.2le", l, x);

       
}


#define LENGTH_LARGEINT 8
char buf_li1[LENGTH_LARGEINT+1] ;
char buf_li2[LENGTH_LARGEINT+1] ;
char buf_li3[LENGTH_LARGEINT+1] ;
char *li1(hcount x) { sprintf_largeint(buf_li1, x, LENGTH_LARGEINT); return buf_li1 ; }
char *li2(hcount x) { sprintf_largeint(buf_li2, x, LENGTH_LARGEINT); return buf_li2 ; }
char *li3(hcount x) { sprintf_largeint(buf_li3, x, LENGTH_LARGEINT); return buf_li3 ; }


void printf_plot(hcount *data, int length,
		 float from, float to) // from et to juste pour l'affichage de la deuxième échelle
{
	hcount min = 0 ;
	hcount max = 0 ;
	hcount tot = 0 ;
	
	float f ;
	int i ;

	for (i=0; i<PLOT_WIDTH; i++)
		datap[i] = 0 ;

	for (i=0; i<length; i++)
	{
		// printf(" %d  %.3lf   \n", i, data[i]);

		if (data[i] < min)
			min = data[i];

		if (data[i] > max)
			max = data[i];

		tot += data[i] ;

		int ii = (int) (i / (float) length * PLOT_WIDTH) ;
		datap[ii] +=  data[i] ;
		datan[ii]++ ;
	}

	hcount oldmax= max ; 
	max = 0 ;

	for (i=0; i<PLOT_WIDTH; i++)
		if (datan[i])
		{
			datap[i] /= datan[i] ;
			if (datap[i] > max)
				max = datap[i] ;
		}
	
	double stepW = ((double) length) / PLOT_WIDTH ;
	double stepH = ((double) (max - min)) / PLOT_HEIGTH ;

	printf("%s (max: %s) = %s elements   ", li1(max), li2(oldmax), li3(tot));
	printf("   first: %.2lf    last: %.2lf  \n", data[0], data[length-1]);
	int h ;
	for (h=PLOT_HEIGTH-1; h>=0; h--)
	{
		hcount b = min + h*stepH ;
		printf("%s ", li1(b));
 
		for (i=0; i<PLOT_WIDTH; i++)
			printf(datap[i] > h*stepH ? "*" : ".");
		printf("\n");
	}


	printf("");
	for (f=0; f<length+stepW; f+=stepW * 10)		
		printf("%10s", "|");
	printf("\n");
			
	printf("");
	for (f=0; f<length+stepW; f+=stepW * 10)		
		printf("%10d", (int) f);
	printf("\n");

	printf("");
	for (f=0; f<length+stepW; f+=stepW * 10)		
		printf("%10.1f", from + ((int) f) * (to - from) / length);
	printf("\n\n");

}




struct st_result 
{
  double from ;
  double to ;
  double step ;
  double err ;

  double from2 ;
  double to2 ;
  double step2 ;
  double err2 ;

  int buckets ;     // size of one dimension
  int all_buckets ; // total size
  hcount *counts ;
} 
;

typedef struct st_result *result ;


void printf_result(result r)
{
	if (!conf_print_plots)
		return ;

        printf("    step  %.3lf   err  %.3lf   err/step   %.3lf\n", r->step, r->err, r->err/r->step);
#ifdef COMPARE
        printf("    step2 %.3lf   err2 %.3lf   err2/step2 %.3lf\n", r->step2, r->err2, r->err2/r->step2);
#endif
        printf_plot(r->counts, r->all_buckets, r->from, r->to);
}




result malloc_result(int buckets)
{
        int all_buckets = buckets ;
#ifdef COMPARE
	all_buckets *= buckets ;
#endif

        result r = (result) malloc (sizeof (struct st_result));
        r -> all_buckets = all_buckets ;
	r -> buckets = buckets ; 
        r -> counts = (hcount *) malloc(r->all_buckets * sizeof(hcount));

	if (r -> counts <= 0)
	{
	        perror("Error in malloc_result");
	        exit(9);
	}
	return r ;
}

void zero_result(result r)
{
        int s ;
	for (s = r->all_buckets -1; s >= 0; s--) 
	        r-> counts[s] = 0 ;
}


void div_result(result r, int d)
{
        int s ;
	for (s = r->all_buckets -1; s >= 0; s--) 
	        r-> counts[s] /= d ;
}

void init_result(result r)
{
	r -> step = (r->to - r->from) / r->buckets ;
	r -> err = r -> step ;
	r -> step2 = (r->to2 - r->from2) / r->buckets ;
	r -> err2 = r -> step2 ;
}

void init_sum_result(result r, result a, result b)
{
        r -> from = a -> from + b -> from ;
	r -> to = a -> to + b -> to ;
        r -> from2 = a -> from2 + b -> from2 ;
	r -> to2 = a -> to2 + b -> to2 ;
	init_result(r);
	r -> err = r -> step + a -> err + b -> err ;
}




// TODO: depassement entier ici ?
hcount two_results_score_to_nb(result resultsA, result resultsB,
					  double score, double err)
{
	int a = resultsA->buckets - 1 ;
	hcount sum_a = 0 ;
	
	int b = 0 ;
	hcount nb = 0 ;
	     
	while (a && (b < resultsB->buckets))
	{
		double score_c = resultsA->from + a*resultsA->step + err*resultsA->err + 
	                         resultsB->from + b*resultsB->step + err*resultsB->err ;

		// printf("  *** %5.3lf \n", score_c);
		if (score_c > score)
		{
			sum_a += resultsA->counts[a] ;
			a-- ;
		}
		else
		{
			nb += sum_a * resultsB->counts[b] ;
			b++ ;
		}
		
		// printf("\n a:%d   b:%d   nb:%.3lf     %.3lf ", a, b, nb, score_c);
	}
	nb += sum_a * resultsB->counts[b] ;
        // printf(" \n %.1f score:%.3lf   nb:%.2lf  \n", err, score, nb);
	
	return nb ;
}
 

void combine_two_results_on_pvalue(result resultsA, result resultsB,
				   hcount pvalue)
{

	printf("\n    SCORE_TO_PVALUE (combining two distributions)\n");

	/*
	double s = resultsA->to + resultsB->to ;
	for ( ; s > -900; s -= 100)
		printf("  sc   %f     nb0  %15.2lf  \n", s, two_results_score_to_nb(resultsA, resultsB, s, 0));
	*/


	// efficiently combine the scores

	double score_bot = resultsA->from + resultsB->from ;
	double score_top = resultsA->to + resultsB->to ;
	int iter = 0 ;


      
	double best_approx = 9999 ;
	double best_score = 0 ;
	int best_iter = -1 ;
	
	while (++iter < 20)
	{
		
		double score_mid = (score_top + score_bot) / 2  ;

		hcount nb0 = two_results_score_to_nb(resultsA, resultsB, score_mid, 0);
		hcount nb1 = two_results_score_to_nb(resultsA, resultsB, score_mid, 1);

		
		hcount dist_pvalue = pvalue - nb0 > nb1 - pvalue ? pvalue - nb0 : pvalue - nb1 ;

		double approx =  ((double) dist_pvalue) / ((double) pvalue);

		if (conf_verbose)
		{
			printf("iter %2d     %10.8lg %10.8lg %10.8lg ", iter, score_bot, score_mid, score_top);
			printf("  nb0 %15.8lg   nb1 %15.8lg ", nb0, nb1);
			printf(" dist %15.8lg words, %6.3lf relative ", dist_pvalue, approx);
		}

		if (ABS(approx) < best_approx)
		{
			if (conf_verbose)
				printf("*");
			best_iter = iter ;
			best_approx = ABS(approx);
			best_score = score_mid ;
		}

		if (conf_verbose)
			printf("\n");


		// printf(" error! %d words, %.3lf relative pvalue\n", nb1-nb0, ((double) (nb1-nb0))/ ((double) pvalue) );

		if (nb1 > pvalue)
			score_bot = score_mid ;
		else if (nb0 < pvalue)
			score_top = score_mid ;
		else
		{
			if (conf_verbose)
				printf(" !! no better approximation \n");
			break ;
		}
	}
	
	printf("      >> best approximation at iter %d\n", best_iter);
	printf("    >>>> use score %.3f\n", best_score);
	printf("    >>>> error = %.3lf relative pvalue  \n", best_approx);
	printf("\n");
}


#define ALLBUCKETS_AB (BIG_NB_BUCKETS)

#ifdef CUFFT
#include <cufft.h>


__global__ void multiply_pointwise(cufftComplex *all, cufftComplex *A, cufftComplex *B, int length, float scale)
{
  // 1 thread
  int i ;
  for (i=0; i<length; i++)
    { 
      all[i].x = (A[i].x * B[i].x - A[i].y * B[i].y) / scale ; 
      all[i].y = (A[i].y * B[i].x + A[i].x * B[i].y) / scale ; 
    }
}

cufftComplex* complex_and_pad(hcount *data, int n)
{
	cufftComplex *out = (cufftComplex *) malloc (2*n * sizeof(cufftComplex));

	int i ;

	for (i=0; i<n; i++)
	{
		out[i].x = data[i] ;
		out[i].y = 0 ;

		out[i+n].x = out[i+n].y = 0;
	}

	return out ;
}

void decomplex(hcount *out, cufftComplex *data, int n)
{
	int i ;
	for (i=0; i<n; i++)
	{
		out[i] = (hcount) (data[2*i].x + data[2*i+1].x) / 2;
	}
}



#endif


result combine_two_results(result resultsA, result resultsB)
{
        printf("      ** Combining two slices\n");

	result all = malloc_result(
#ifdef CUFFT
		2*
#endif 
		NB_BUCKETS);
 
	init_sum_result(all, resultsA, resultsB) ;
	zero_result(all);
 
#ifdef CUFFT
	// assert(resultsA -> all_buckets == resultsB -> all_buckets);
	int n = resultsA -> all_buckets ;

	cufftHandle plan ;
	printf("plan::%d\n", cufftPlan1d(&plan, 2*n, CUFFT_C2C, 1));

	// A
	cufftComplex *AC = complex_and_pad(resultsA -> counts, n);
	cufftComplex *device_AC ; 
	cudaMalloc((void**) &device_AC, 2*n * sizeof(cufftComplex)); 
	cudaMemcpy(device_AC, AC, 2*n * sizeof(cufftComplex), cudaMemcpyHostToDevice); 
 	printf(" (A):%d\n", cufftExecC2C(plan, device_AC, device_AC, CUFFT_FORWARD));

	// B
	cufftComplex *BC = complex_and_pad(resultsB -> counts, n);
	cufftComplex *device_BC ; 
	cudaMalloc((void**) &device_BC, 2*n * sizeof(cufftComplex)); 
	cudaMemcpy(device_BC, BC, 2*n * sizeof(cufftComplex), cudaMemcpyHostToDevice); 
 	printf(" (B):%d\n", cufftExecC2C(plan, device_BC, device_BC, CUFFT_FORWARD));

	// 
	dim3  grid(1, 1, 1);
	dim3 threads(1, 1, 1);
	cufftComplex *device_allC ;
	cudaMalloc((void**) &device_allC, n * sizeof(cufftComplex)); 
	printf("multiply pointwise\n");
	multiply_pointwise <<< grid, threads>>>(device_allC, device_AC, device_BC, 2*n, n); 
	
	printf(" (=):%d\n", cufftExecC2C(plan, device_allC, device_allC, CUFFT_INVERSE));

	cufftDestroy(plan);

	cudaMemcpy( AC, device_allC, 2*n * sizeof(cufftComplex), cudaMemcpyDeviceToHost); 
	decomplex(all -> counts, AC, 2*n );
	free(AC);
	free(BC);

	// printf_result(all);
	all -> all_buckets /= 2 ;

#else // CUFFT

	int sA, sB ;
	for (sA = resultsA->all_buckets-1; sA >= 0; sA--) 
		for (sB = resultsB->all_buckets-1; sB >= 0; sB--) 		
		{
#ifdef COMPARE
         	        int cA = sA % NB_BUCKETS ;
			int dA = sA / NB_BUCKETS ;
         	        int cB = sB % NB_BUCKETS ;
			int dB = sB / NB_BUCKETS ;

			int sc = (int) ((((double) cA) * resultsA->step +
					 ((double) cB) * resultsB->step)
					/ all->step) ;

			int sd = (int) ((((double) dA) * resultsA->step2 +
					 ((double) dB) * resultsB->step2)
					/ all->step2) ;

			int s = sc + sd * NB_BUCKETS ;
			// printf(" %d %d  -> %d \n", sA, sB, s);
#else
			int s = (int) ((((double) sA) * resultsA->step +
					((double) sB) * resultsB->step)
					       / all->step) ;
#endif
			all->counts[s] += resultsA->counts[sA] * resultsB->counts[sB] ;
		}


#endif // CUFFT

	// plot combined score
	printf_result(all);

	return all ;

}



result combine_n_results(result *results, int start, int end)
{
        printf("    **** Combining slices %d to %d\n", start, end);
        
	result old = results[start] ;
	result ner = old ;

	int i ;
        for (i=start+1; i<= end; i++)
        {
	   ner = combine_two_results(old, results[i]);
	   old = ner ;
	}

 	// free old
	return ner ;
}

				  
#define EPSILON 1e-10  // pour erreurs d'arrondi sur le score


void compute_score_from_pvalue(result all, hcount pvalue)
{
	hcount tot = 0 ;
	int s ;
	for (s = all->buckets-1; s >= 0; s--) 
	{
		tot += all->counts[s] ;
		// printf (" %d  %.3lf \n", s, tot); 
		if (tot >= pvalue)
			break ;
	}

	int start_s = s ;
	float start_score = all->from + all->step * s ;

	printf("\n    PVALUE_TO_SCORE\n");
	printf("      >> s = %d   score = %.3f \n", s, start_score);

	int err_buck = (int) ceil(all->err/all->step - EPSILON) ;
	s += err_buck ;

	int stop_s = s ;
	float stop_score = all->from + all->step * s ;
	
	printf("      >> s = %d   score = %.3f \n", s, stop_score);
			
	tot = 0 ;
	for (s = start_s; s < stop_s; s++) 
                tot += all->counts[s] ;

	printf("    >>>> use score %.3f\n", (start_score + stop_score) / 2);
	printf("    >>>> error = %.3f score (to %.3f), %d buckets, %.8lg words >>>> %.3lf relative pvalue  \n",
	       all-> err,  all->from + all->step * start_s + all->err, err_buck,  tot, (double) tot / (double) pvalue);
	
}

void compute_pvalue_from_score(result all, float score, int length)
{
	hcount tot = 0 ;
	int s ;

	int start_s = (int) ((score - all->from) / all->step) ;
	for (s = all->buckets-1; s >= start_s; s--) 
		tot += all->counts[s] ;


	float start_score = all->from + all->step * s ;


	printf("\n    SCORE_TO_PVALUE\n");
	printf("      >> s = %d   score = %.3f   words = %.8lg \n", s, start_score, tot);

	int err_buck = (int) ceil(all->err/all->step - EPSILON) ;
	s += err_buck ;

	int stop_s = s ;
	float stop_score = all->from + all->step * s ;
	
	printf("      >> s = %d   score = %.3f \n", s, stop_score);
			
	hcount start_tot = tot ;
	
	for (s = start_s; s < stop_s; s++) 
                tot += all->counts[s] ;

	hcount stop_tot = tot ;
	hcount words = (start_tot + stop_tot) / 2;
	float pvalue = words / pow(4.0, length) ;
	hcount err_words = stop_tot - start_tot ;

	printf("    >>>> use words %.8lg, pvalue %.3e\n", words, pvalue);
	printf("    >>>> error = %.3f score (to %.3f), %d buckets, %.8lg words >>>> %.3lf relative pvalue  \n",
	       all-> err,  stop_score, err_buck,  err_words, (double) err_words / (double) stop_tot);
	
}
