/*
 * CUDA PWM - Mathieu Giraud & Jean-Stéphane Varré, 2008
 */

#include <stdio.h>
#include <math.h>


__device__ void printf_word(int *word, int length)
{
  int t ;
  for (t=0; t<length; t++)
    PRINTF("%c", "ACGTacgt"[word[t]]);
  
  PRINTF(" ");
}

void printf_matrix(float *matrix, int length)
{
  int i ;

  if (conf_verbose)
  for (i=0; i<ALPH_SIZE*length; i++)
    {
      printf(" %8.3f ", matrix[i]);
      if (!((i+1)%ALPH_SIZE))
	printf("\n");

      if (i==8*ALPH_SIZE+1)
      { 
	      printf ("    ...\n");
	      break ;
      }
    }

  printf("== <-- length: %d\n\n", length);
}

void printf_seq(unsigned char *seq, int length)
{
  int i ;
  for (i=0; i<length; i++)
    {
      printf("%d", (int) seq[i]);
    }

  // printf("\n== <-- length: %d\n\n", length);
}

float *random_matrix(int length)
{
  printf(" <-- random matrix\n");

  float *matrix = (float *) malloc(ALPH_SIZE * length * sizeof(float));

  int i ;
  for (i=0; i<ALPH_SIZE*length; i++)
    matrix[i] = (1 << 2) - ((float) random()) / (1 << 28) ;

  printf_matrix(matrix, length);

  return matrix ;
}


unsigned char *random_seq(int length)
{
  printf(" <-- random seq: %.3e\n", (float) length);

  unsigned char *seq = (unsigned char*) malloc(sizeof(unsigned char) * length);

  int i ;
  for (i=0; i<length; i++)
    seq[i] = (random() % ALPH_SIZE) ;

  // printf_seq(seq, length);

  return seq ;
}

unsigned char *seq_from_file(const char *fname, int *length)
{
  printf(" <-- %s\n", fname);

	int code[256] ;
	int i ;
	for (i=0; i<255; i++)
		code[i] = 9 ;

	code['A'] = code['a'] = 0 ;
	code['C'] = code['c'] = 1 ;
	code['G'] = code['g'] = 2 ;
	code['T'] = code['t'] = 3 ;


	FILE* f = fopen(fname, "r");
	if (f == 0)
	{
		perror(fname);
		exit(79);
	}

  unsigned char *seq = (unsigned char*) malloc(sizeof(unsigned char) * 500000000);
	char c ;
	*length = 0 ;

  while ((c = fgetc(f)) != EOF)
	{
		int n = code[c];
		if (n == 9)
			continue ;

		seq[(*length)++] = n;
	}

	printf(" <-- length: %d\n", *length);

  return seq ;
}


float matrix_buf[ALPH_SIZE*MATRIX_MAX_LENGTH];

float *matrix_from_file(float* matrix_force, const char *fname, int *length)
{
  printf("== <-- %s\n", fname);
  FILE* f = fopen(fname, "r");

  if (f == 0)
    {
      perror (fname);
      exit(5);
    }


  // Read in a buffer
  int n = 0 ;
  int stop = 0 ;

  while (stop != 1)
    {
      int a ;
      for (a=0; a<ALPH_SIZE; a++)
	{
	  if (fscanf(f, "%f", matrix_buf+n) <= 0)
	    { 
	      if (a+n+stop == 0)
		{
		  stop = 2 ;
		  if (fscanf(f, "> %s\n"), (char *) matrix_buf <=0 )
		    break ;

		  printf (">>> %s\n", (char *) matrix_buf);
		  a-- ;
		  continue ;
		}
	     
	      
	      stop = 1 ;
	      break ;
	    }
	  n++ ;
	}
    }

  *length = n/4 ;

  // Prepare new matrix
  float *matrix = matrix_force ? matrix_force : (float *) malloc(ALPH_SIZE * (*length) * sizeof(float));

  int a ;
  for (n=0; n<*length; n++)
    for (a=0; a<ALPH_SIZE; a++)
      matrix[n*ALPH_SIZE+a] = matrix_buf[a*(*length)+n] ;

  printf_matrix(matrix, *length);

  return matrix; 
}


void display_results(int *results, int *results_pos, int length)
{
  long  long  nbocc = 0 ;
  int i;
  int results_max = 0 ;

  int bad_threads = 0 ;

  for (i=0; i<NB_THREADS_SCAN; i++)
    {
      // printf(" %2d", results[i]);
      nbocc += results[i] ;

      if (results[i] > RES_PER_THREAD)
	{
	  bad_threads++ ;
	  if (bad_threads < 5)
	    printf("  Warning: thread %d has %d results \n", i, results[i]);
	}

      if (results[i] > results_max)
        results_max = results[i];
    }

  if (bad_threads >= 5)
    printf ("  Warning: .... %d threads with too many results.\n  Please use a higher threshold or compile with an higher RES_PER_THREAD.\n", bad_threads);
    
  
  printf(">>> %lld occurences ", nbocc);
  printf("(best thread: %d) ", results_max);

#ifdef OLD_PVALUE
  printf(">>> p-value: %.6e\n", (float) nbocc / pow((float) ALPH_SIZE, (float) length)); 
#endif

  printf(">>> prob: %.6e\n", (float) nbocc / (float) length);
}



#define ABS(a)  ((a>0)?(a):-(a))
#define MAX(a,b) ((a)>(b)?(a):(b))
#define MIN(a,b) ((a)>(b)?(b):(a))

__device__ __host__ float max_score_column (float *column) {
	return MAX(MAX(*column, *(column+1)),MAX(*(column+2),*(column+3)));
}

__device__ __host__ float min_score_column (float *column) {
	return MIN(MIN(*column, *(column+1)),MIN(*(column+2),*(column+3)));
}

__device__ __host__ float max_score (float *matrix, int length) {
	float sum = 0;      
	for (int i = 0; i < length; i++) {
		sum += max_score_column (matrix + i * ALPH_SIZE);
	}
	return sum + 1 ;
}

__device__ __host__ float min_score (float *matrix, int length) {
	float sum = 0;      
	for (int i = 0; i < length; i++) {
		sum += min_score_column (matrix + i * ALPH_SIZE);
	}
	return sum;
}
