/*
 */



// a mettre dans pwm.h

#define SLICE_LENGTH(s) (((s) == (nbSlices-1) && supplSliceSize) ? supplSliceLength : SLICEWIDTH)
#define SLICE_SIZE(s) (((s) == (nbSlices-1) && supplSliceSize) ? supplSliceSize : sliceSize)


__device__ int pwm_score(const float* matrix, const float* group_matrix,
			 int* word, int length, float threshold, 
												 float* best_scores, float *worst_scores,
												 int nbSlices, int supplSliceLength, long sliceSize, long supplSliceSize)
{

  float score = 0 ;

	int *w = word;

#ifdef SLICES_AUDE

	int slice;
	for (slice = 0; slice < nbSlices; slice++) {
		// w += 4;
		//	int intword = 4 * 4 * 4 * (*(w-3)) + 4 * 4 * (*(w-2)) + 4 * (*(w-1)) + *w;
		w += 2;
		int intword =  4 * (*(w-1)) + *w;
		/*
		int intword = 0;
		int pos ;
		for (pos = 0; pos < SLICE_LENGTH(slice); pos++) {
			//PRINTF("%1d",*w);
			intword = intword * ALPH_SIZE + (*w);
			w++;
			}
		*/
		//PRINTF(" %d\n",intword);
		score += group_matrix[slice * SLICESIZE + intword];
	}

#else

	int t;
  for (t=0; t<length; t++)
    score += matrix[ALPH_SIZE*t + (*w++)] ;

#endif

// OLD VERSION
/*

  int *w = word ;
  int t;

#ifdef SLICES_AUDE

  for (t=0; t<INDEX_NO_WU/2; t++)
    {
      int a = *w++ ;
      int b = *w++ ;
      score += group_matrix[ALPH_SIZE*ALPH_SIZE*t + ALPH_SIZE*a + b] ;
    }

  for (; t<length/2; t++)
    {
      int a = *w++ ;
      int b = *w++ ;
      score += group_matrix[ALPH_SIZE*ALPH_SIZE*t + ALPH_SIZE*a + b] ;

#else // SLICES_AUDE

  for (t=0; t<INDEX_NO_WU; t++)
    score += matrix[ALPH_SIZE*t + (*w++)] ;

  for (; t<length; t++)
    {
      score += matrix[ALPH_SIZE*t + (*w++)] ;

#endif


#ifdef WU_BEST
      if (score + best_scores[t+1] < threshold)
	return 0 ;
#endif
     
#ifdef WU_WORST
     if (score + worst_scores[t+1] >= threshold)
	return 1 ;
#endif

    }

  // (Unused if WU_WORST)
	*/
  

  DEBUG printf_word(word, length);
  DEBUG PRINTF("  %5.8f \n", score);

  return (score >= threshold) ;
}



__device__ void group_matrix_mathieu(float *group, float *matrix, int length)
{
  int v = 0 ;
  int a, b, i; 

  // to parallelize, operations dupliquees pour l'instant...
  for (i=0; i<length/2; i++)
    for (a=0; a<ALPH_SIZE; a++)
      for (b=0; b<ALPH_SIZE; b++)
	group[v++] = matrix[ALPH_SIZE*(2*i) + a] + matrix[ALPH_SIZE*(2*i+1) + b] ;
}


// B_JS


__device__ void group_matrix(float *t, float *matrix, int length, int nbSlices, int supplSliceLength, long sliceSize, long supplSliceSize) 
{

	
	float *toffset = t;
  for (int slice = 0; slice < nbSlices; slice++) {
    for (int i = 0; i < SLICE_SIZE(slice); i++) {
			(*toffset) = 0;
			toffset++;
		}
  }
	

//	PRINTF("%2d %2d - %2d %2d - %.0f %.0f\n",length, SLICE_WIDTH, nbSlices, supplSliceLength, sliceSize, supplSliceSize);
  int l = 0;
  while (l < SLICEWIDTH) { // must be maximal slice length
    long nbSpan = pow(4.0,1.0*l); // number of successive cases with same score to be added
    long nbStep = nbSpan * ALPH_SIZE;    // increasing step
//		PRINTF("span = %ld step %ld\n",nbSpan, nbStep);
		for (int slice = 0; slice < nbSlices; slice++) {
			if (l < SLICE_LENGTH(slice)) {
				for (int x = 0; x < ALPH_SIZE; x++) {
					for (int step = 0; step < SLICE_SIZE(slice); step += nbStep) {
						for (int span = 0; span < nbSpan; span++) {
              t[slice * SLICE_SIZE(0) + x*nbSpan+step+span] += matrix[slice * SLICEWIDTH * ALPH_SIZE + l * ALPH_SIZE + x];              
//              PRINTF("slice = %2d matrix[%2d] = %0.f  t[%2d] = %.0f\n", slice, slice * SLICEWIDTH * ALPH_SIZE + l * ALPH_SIZE + x , matrix[slice * SLICEWIDTH * ALPH_SIZE + l * ALPH_SIZE + x], slice * SLICE_SIZE(0) + x*nbSpan+step+span,t[slice * SLICE_SIZE(0) + x*nbSpan+step+span] );
            }
          }
        }
      }
      //if (slice == 0) cout << endl;
    }
    
    l++;
	}
  
	
  // DEBUG
	int s;
	for (s = 0; s < nbSlices; s++) {
		PRINTF("slice %2d : [ ",s);
		int w;
		for (w = 0; w < SLICE_SIZE(s); w++) {
			PRINTF("%.0f ",t[s*sliceSize+w]);
		}
		PRINTF("]\n");
	}
	

}
// E_JS

__device__ void compute_best_worst_scores(float *matrix, int length, float *best_scores, float *worst_scores)
{
  // temps: +1%
  // to parallelize, operations dupliquees pour l'instant...

    int v = (length+1) * ALPH_SIZE - 1;
    int l;
    best_scores[length+1] = worst_scores[length+1] = 0 ;

    for (l=length; l>=0; l--)
      { 
	float best = matrix[v] ;
	float worst = matrix[v] ;
	
	v-- ;

	int a ;
	for (a=1; a<ALPH_SIZE; a++)
	  {
	    if (matrix[v] > best)
	      best = matrix[v] ;
	    
	    if (matrix[v] < worst)
	      worst = matrix[v] ;
	
	    v-- ;
	  }

	best_scores[l] = best_scores[l+1] + best ;
	worst_scores[l] = worst_scores[l+1] + worst ;
	
	// DEBUG PRINTF(" l:%d   best:%.4f  worst:%.4f  \n", l, best_scores[l], worst_scores[l]);
      }
 
#ifdef SLICES_AUDE
    for (l=length; l>=0; l--)
      {
	best_scores[2*l] = best_scores[l];
	worst_scores[2*l] = worst_scores[l];
      }
#endif
}

  

__global__ void scan_pwm (float *matrix_device, int length, float threshold, int* results, int* results_pos, unsigned char* seq_device, int seq_length)
{

  // Copy the matrix
  __shared__ float matrix_shared[ALPH_SIZE * MATRIX_MAX_LENGTH_SCAN] ;
  copy_float_device_shared(matrix_shared, matrix_device, length*ALPH_SIZE);
   
  // Group the matrix 2/2

	// B_JS
	int nbSlices = length / SLICEWIDTH;
	int supplSliceLength = 0;
	supplSliceLength = length % SLICEWIDTH;
	if (supplSliceLength > 0) nbSlices++;

	long sliceSize = SLICESIZE;
	long supplSliceSize = 0; 
	if (supplSliceLength > 0) supplSliceSize = pow(4.0,1.0*supplSliceLength);

  __shared__ float group_matrix_shared[((MATRIX_MAX_LENGTH_SCAN / SLICEWIDTH)+1) * SLICESIZE];

	// E_JS
#ifdef SLICES_AUDE
//  group_matrix_mathieu(group_matrix_shared, matrix_shared, length);
  group_matrix(group_matrix_shared, matrix_shared, length, nbSlices, supplSliceLength, sliceSize, supplSliceSize);
#endif
 
  // Compute best/worst scores

  __shared__ float best_scores[MATRIX_MAX_LENGTH_SCAN] ;
  __shared__ float worst_scores[MATRIX_MAX_LENGTH_SCAN] ;

#ifdef WU_BEST
  compute_best_worst_scores(matrix_shared, length, best_scores, worst_scores);
#endif

  // Determine my id
  const unsigned int bx = blockIdx.x;
  const unsigned int tx = threadIdx.x;
  long id = bx * NB_THREADS_PER_BLOCK_SCAN + tx ;
  // PRINTF("Block %5d, thread %5d, id %6ld ", bx, tx, id);

  int *results_out = results_pos + id * RES_PER_THREAD ;

  // Copy the seq
  // long slices_per_block = 1 ;
  long slices_per_block = seq_length / (NB_BLOCKS_SCAN * SEQ_LENGTH_SHARED)  ;

  __shared__ int seq_shared[SEQ_LENGTH_SHARED + MATRIX_MAX_LENGTH_SCAN] ;    

//  int t; for (t=0; t<16; t++) // for benchmarks
  {


  int nbocc = 0 ;
  int *seq ;
  int slice = 0;


  for (slice = 0; slice < slices_per_block; slice++)
    {
       copy_seq_device_shared(seq_shared, seq_device + SEQ_LENGTH_SHARED * (bx*slices_per_block + slice), SEQ_LENGTH_SHARED + MATRIX_MAX_LENGTH_SCAN);

       // __syncthreads();


#ifdef THREADS_INTERLEAVED
       // seq = seq_shared + tx;
       for (seq = seq_shared + tx; seq < seq_shared + SEQ_LENGTH_SHARED; seq+= NB_THREADS_PER_BLOCK_SCAN)
#else
       int SEQ_LENGTH_BY_THREAD = SEQ_LENGTH_SHARED / NB_THREADS_PER_BLOCK_SCAN ;
       for (seq = seq_shared + tx * SEQ_LENGTH_BY_THREAD; 
	    seq < seq_shared + (tx+1) * SEQ_LENGTH_BY_THREAD; 
	    seq+= 1)
#endif
         {
	   int good = pwm_score(matrix_shared, group_matrix_shared,
				seq, length, threshold, 
													best_scores, worst_scores,
													nbSlices, supplSliceLength, sliceSize, supplSliceSize);  
   
           if (good)
             {
	       nbocc++ ;
	       results_out[nbocc % RES_PER_THREAD] = (seq-seq_shared) ; 
             }
         }
    }
  DEBUG printf_word(seq, length);
  DEBUG PRINTF("  --> %5d occurences\n", nbocc);

  __syncthreads();

  results[id] = nbocc ;
}  
}



void scan(float *matrices, int nb_matrices, int *lengths,  unsigned char *seq, int seq_length, float threshold)
{
  printf("== Seq (%d) : ", seq_length) ;
  printf_seq(seq, 50);
  printf("...\n");

  // allocate device memory for matrices and copy them
  float *device_matrices; 
  cudaMalloc((void**) &device_matrices, nb_matrices * ALPH_SIZE * MATRIX_MAX_LENGTH * sizeof(float)); 
  cudaMemcpy(device_matrices, matrices, nb_matrices * ALPH_SIZE * MATRIX_MAX_LENGTH * sizeof(float), cudaMemcpyHostToDevice);  

  // allocate device memory for seq and copy them
  unsigned char *device_seq; 
  cudaMalloc((void**) &device_seq, seq_length + MATRIX_MAX_LENGTH); 
  cudaMemcpy(device_seq, seq, seq_length, cudaMemcpyHostToDevice);  

  // circular padding
  cudaMemcpy(device_seq + seq_length, seq, MATRIX_MAX_LENGTH, cudaMemcpyHostToDevice);  
  int *device_results_pos; 
  cudaMalloc((void**) &device_results_pos, NB_THREADS_SCAN * sizeof(float) * RES_PER_THREAD); 
  int *results_pos = (int *) malloc(NB_THREADS_SCAN * sizeof(int) * RES_PER_THREAD); 
  
  // allocate device memory for results
  int *device_results; 
  cudaMalloc((void**) &device_results, NB_THREADS_SCAN * sizeof(float)); 
  int *results = (int *) malloc(NB_THREADS_SCAN * sizeof(int)); 
  
  // execute the kernel
  dim3  grid(NB_BLOCKS_SCAN, 1, 1);
  dim3  threads(NB_THREADS_PER_BLOCK_SCAN, 1, 1);
  printf("== Blocks: %4d   Threads/block: %4d\n", NB_BLOCKS_SCAN, NB_THREADS_PER_BLOCK_SCAN);

 
  int n ; 

	float *device_matrix = device_matrices ; 

	for(n=0; n<nb_matrices; n++)
	{
		time_stop();
		printf("[matrix:%d] \n", n);

		scan_pwm <<< grid, threads >>> (device_matrix, lengths[n], threshold, device_results, device_results_pos, device_seq, seq_length);
		CUT_CHECK_ERROR("kernel 'scan' failed");

		device_matrix += ALPH_SIZE * lengths[n] ;
		// copy the results
		cudaMemcpy(results, device_results, NB_THREADS_SCAN * sizeof(float), cudaMemcpyDeviceToHost);  
		cudaMemcpy(results_pos, device_results_pos, NB_THREADS_SCAN * sizeof(float) * RES_PER_THREAD, cudaMemcpyDeviceToHost);  
		
		display_results(results, results_pos, seq_length);
	}
}






