


#define POS(i)  ( ((i)*length)/N )
#define LENGTH(i)  ( POS(i+1) - POS(i) )
#define LENGTH_THREAD(i)  (LENGTH(i) - (LOG_GRID_X + LOG_NB_THREADS_PER_BLOCK) / LOG_ALPH_SIZE)


#define MATRIX_OFFSET(i)  ( ALPH_SIZE * POS(i) )
#define RESULTS_OFFSET(i)  ( BIG_NB_BUCKETS * (i) )




__device__ void from_to_step(float *matrix, int length, float *from, float *to, float *step, int by, int bz)
{
	*from = min_score(matrix, length);
	*to = max_score(matrix, length);

	*step = (*to - *from) / NB_GRID_Y ;
	*from += by * *step ;
	*to = *from + *step ; 
	*step /= NB_BUCKETS ;
}


__global__ void score_buckets (float *matrix_device,
#ifdef COMPARE
				float *matrix_device2,		
#endif
				int length, int N,
				 dcount* results)

{
	// Determine my id
	const unsigned int bx = blockIdx.x; // enumeration
	const unsigned int by = blockIdx.y / NB_GRID_Z; // precision
	const unsigned int bz = blockIdx.y % NB_GRID_Z; // matrix slice

	int length_slice  = LENGTH(bz) ; 
	int length_thread = LENGTH_THREAD(bz) ;

	// Copy the slice of the matrix
	__shared__ float matrix_shared_slice[ALPH_SIZE * SLICE_MAX_LENGTH] ;
	copy_float_device_shared(matrix_shared_slice, 
				 matrix_device + MATRIX_OFFSET(bz), 
				 ALPH_SIZE*length_slice);

#ifdef COMPARE
	__shared__ float matrix_shared_slice2[ALPH_SIZE * SLICE_MAX_LENGTH] ;
	copy_float_device_shared(matrix_shared_slice2, 
				 matrix_device2 + MATRIX_OFFSET(bz), 
				 ALPH_SIZE*length_slice);
#endif

	const unsigned int tx = threadIdx.x; // enumeration


	// Allocation for results
	__shared__ count results_shared[NB_ALL_BUCKETS] ;
	for (int i = tx; i < NB_ALL_BUCKETS; i+= NB_THREADS_PER_BLOCK)
	{
		results_shared[i] = 0 ;
	}
	__syncthreads();

       
	// enumeration
	long id = tx + bx * NB_THREADS_PER_BLOCK ; // enumeration 

	// precision : from/to/step
	float from, to, step ;
	from_to_step(matrix_shared_slice, length_slice, &from, &to, &step, by, bz);

#ifdef COMPARE
	float from2, to2, step2 ;
	from_to_step(matrix_shared_slice2, length_slice, &from2, &to2, &step2, by, bz);
#endif

	// resultats:  precision + slice
	dcount *results_slice = results + RESULTS_OFFSET(bx*N + bz) + by * NB_ALL_BUCKETS ;
	

	// Initialize word, depending on id
	int word[SLICE_MAX_LENGTH] ;
	generate_word(word, 0, length_thread + 1 );  
	generate_word(word, 0, length_slice);  
	generate_word(word + length_thread, id, length_slice - length_thread);  

	// Intialize score vectors, depending on id
	float score[SLICE_MAX_LENGTH+1];   
	float sum_score = initialize_score(matrix_shared_slice, score, word, length_thread, length_slice);

#ifdef COMPARE
	float score2[SLICE_MAX_LENGTH+1];   
	float sum_score2 = initialize_score(matrix_shared_slice2, score2, word, length_thread, length_slice);
#endif

	int p = 0;

	// PRINTF("I'm thread %d  block %d %d %d      %d ", tx, bx, by, bz,  id);
	// PRINTF("length_slice %d   length_thread %d \n", length_slice, length_thread);
	
	do {
		sum_score = change_score(matrix_shared_slice,score,word,p,sum_score);
#ifdef COMPARE
		sum_score2 = change_score(matrix_shared_slice2,score2,word,p,sum_score2);
#endif
		
		if (sum_score >= from && sum_score <= to) 
		{
			int cell = floor((sum_score - from) / step) ;

#ifdef COMPARE
			cell += NB_BUCKETS * floor((sum_score2 - from2) / step2) ;
#endif

#ifdef ATOMICS
			atomicInc((unsigned int*) results_shared + cell, (count) -1);
#else
#ifdef NOSERIAL
			results_shared[cell]++ ;  
#else
			SERIALIZE ( results_shared[cell]++ ; )  ;
#endif
#endif
			// printf_word(word, length_slice) ;
			// PRINTF(" %.3f   %d    %.3f  %.3f  .%3f \n",  sum_score, cell,  from, to, step);
		}

		p = inc_word(word);  		

	} while (p < length_thread);

	__syncthreads();

	// Save the results to device
	for (int i = tx; i < NB_ALL_BUCKETS; i+= NB_THREADS_PER_BLOCK) {
		results_slice[i] = results_shared[i] ;
	}
}



float *matrix_on_device(float *matrix, int length)
{
	// allocate device memory for matrices and copy them
	if (length >= MATRIX_MAX_LENGTH)
	{
	        printf ("   Error : matrix too large. Please compile with a larger MATRIX_MAX_LENGTH\n");
	        printf("Error:  MATRIX_MAX_LENGTH is too small.\n");
                exit(8);
	}

        float *device_matrix ;
	cudaMalloc((void**) &device_matrix, ALPH_SIZE * MATRIX_MAX_LENGTH * sizeof(float)); 
	cudaMemcpy(device_matrix, matrix, ALPH_SIZE * MATRIX_MAX_LENGTH * sizeof(float), cudaMemcpyHostToDevice); 

	return device_matrix ;
}


void pvalue2score_n (float *matrix, 
#ifdef COMPARE
		     float *matrix2,
#endif
				 int length, int N, int pvalue_exp, float score_goal)
{
	printf("== cudaMalloc/Memcpy...");
	fflush(stdout);

       	float *device_matrix1 = matrix_on_device(matrix, length);

#ifdef COMPARE
	float *device_matrix2 = matrix_on_device(matrix2, length);
#endif

	// allocate device memory for results
	dcount *device_results; 
	cudaMalloc((void**) &device_results, N * NB_GRID_X * BIG_NB_BUCKETS * sizeof(dcount)); 
	dcount *results = (dcount *) malloc(N * NB_GRID_X * BIG_NB_BUCKETS * sizeof(dcount)); 

	printf(" [ok]\n");
	time_stop();

        // allocate and prepare results[N]
	result resultsN[N] ;

	float Tfrom = 0 ;
	float Tto = 0 ;

#ifdef COMPARE
	float Tfrom2 = 0 ;
	float Tto2 = 0 ;
#endif

	int i;
	for (i=0; i<N; i++)
	{
	        resultsN[i] = malloc_result(NB_BUCKETS) ;
		Tfrom += resultsN[i]->from =  min_score(matrix + MATRIX_OFFSET(i), LENGTH(i));
		Tto   += resultsN[i]->to   =  max_score(matrix + MATRIX_OFFSET(i), LENGTH(i));

#ifdef COMPARE
		Tfrom2 += resultsN[i]->from2 =  min_score(matrix2 + MATRIX_OFFSET(i), LENGTH(i));
		Tto2   += resultsN[i]->to2   =  max_score(matrix2 + MATRIX_OFFSET(i), LENGTH(i));
#endif
	      
		init_result(resultsN[i]);
		zero_result(resultsN[i]);

		printf("     [slice %d]  POS: %2d  LENGTH: %2d  From: %.3f     To: %.3f   ", 
		       i, POS(i), LENGTH(i), 
		       resultsN[i]->from, resultsN[i]->to);

#ifdef COMPARE
		printf("     From2: %.3f     To2: %.3f   ", 
		       resultsN[i]->from2, resultsN[i]->to2);
#endif
		int length_thread = LENGTH(i) - (LOG_GRID_X + LOG_NB_THREADS_PER_BLOCK) / LOG_ALPH_SIZE  ; 
		printf("  length_thread: %d \n", length_thread);

		if (length_thread <= 0)
		{
			printf ("   Error : length_thread <= 0. Please use a larger matrix, or a smaller N.\n");
			exit(2);
		}

		if (LENGTH(i) >= SLICE_MAX_LENGTH)
		{
			printf ("   Error : length of the slice too large. Please use a larger N, or compile with a larger SLICE_MAX_LENGTH\n");
			exit(2);
		}

	}
	
	printf("       [total]  POS: %2d              From: %.3f     To: %.3f  \n", 
	       POS(N), Tfrom, Tto);

	

	// execute the kernel
	printf("== Length: %4d   Threads: %4d \n", length, NB_THREADS);
	printf("== Enumeration: %d   Precision: %d  Slices: %d ", 
	       NB_GRID_X, NB_GRID_Y, NB_GRID_Z);
	printf("\n== Grid: %4d   Threads/block: %4d\n", NB_BLOCKS, NB_THREADS_PER_BLOCK);
	dim3  grid(NB_GRID_X, NB_GRID_Y * NB_GRID_Z, 1);
	dim3  threads(NB_THREADS_PER_BLOCK, 1, 1);

	printf("== Launching kernel...");
	fflush(stdout);

	score_buckets <<< grid, threads >>> (device_matrix1, 
#ifdef COMPARE
					      device_matrix2,
#endif
					      length, N, device_results); 

        CUT_CHECK_ERROR("kernel 'score_buckets' failed");
	printf(" [ok]\n");
	time_stop();


	// copy the results
	printf("== Copy and sum results...");
	fflush(stdout);
	cudaMemcpy(results, device_results, 
		   N * NB_GRID_X * BIG_NB_BUCKETS * sizeof(dcount), cudaMemcpyDeviceToHost);  
	

	// sum over GRID_X
	int bx ;
	int j ;

	for (i=0; i<N; i++)
	{
		hcount t = 0 ;

		for (bx=0; bx<NB_GRID_X; bx++)
		{
			for (j=0; j<BIG_NB_BUCKETS; j++)
			{
				hcount r = (hcount) results[RESULTS_OFFSET(bx*N + i) + j] ;
				resultsN[i] -> counts[j] += r ;
				t += r ;
			}
		}


		hcount loss = pow(4.0, LENGTH(i)) - t ;
		float relative_loss = (float) loss / (float) pow(4.0, LENGTH(i));

		if (loss > .5)
			printf ("(loss for slice %d: %8.2lf, %.4f relative) ", i, loss, relative_loss);

		if (relative_loss > .1)
		{
			printf("    !! There is a big problem.\n", i);
			exit(7);
		}

	}


	printf(" [ok]\n");
	time_stop();
	
	// plot the results
	for (i=0; i<N; i++)
	          printf_result(resultsN[i]);


	// Combine the results
	printf("== Combining the results...\n");

	result all = (result) 0;

	hcount pvalue =  (hcount) (pow(4.0,(double) length)
				   /pow(10.0, (double) pvalue_exp)) ;
	
	if (N > 1)
	{
		int half = N/2;
		result fst_half = combine_n_results(resultsN, 0, half-1);
		result snd_half = combine_n_results(resultsN, half, N-1);
		
		if (pvalue_exp)
		        combine_two_results_on_pvalue(fst_half, snd_half, pvalue);
		
		if (conf_last_combine)
			all = combine_two_results(fst_half, snd_half);
	}

	else
		all = resultsN[0] ;


	if (all && pvalue_exp)
	        compute_score_from_pvalue(all, pvalue);

	if (all)
		compute_pvalue_from_score(all, score_goal, length);
		 
	if (pvalue_exp)
	        printf("    >>>> pvalue was %.8lg words (%d)\n", pvalue, pvalue_exp);


	// Loop on all pvalues
	if (all && !pvalue_exp)
	{
		int exp ;

		for (exp = 1; exp <= 10; exp++) 
		{
			printf ("\n");

			hcount pvalue =  (hcount) (pow(4.0,(double)length)
						   /pow(10.0, (double) exp)) ;
			compute_score_from_pvalue(all, pvalue);
			
			printf("    >>>> pvalue was %.8lg words (%d)\n", pvalue, exp);
		}
	}

}



